/* eslint-disable @typescript-eslint/no-unsafe-assignment */
import assert from 'node:assert';
import { camelToSnake, camelKeys, snakeKeys, snakeToCamel, } from '@waiting/shared-core';
import { CaseType, EnumClient, } from './types.js';
export async function getCurrentTime(dbh, clientType) {
    if (clientType) {
        const res = await dbh.raw('SELECT CURRENT_TIMESTAMP AS currenttime;');
        switch (clientType) {
            case EnumClient.pg:
            // eslint-disable-next-line no-fallthrough
            case EnumClient.pgnative:
            // eslint-disable-next-line no-fallthrough
            case EnumClient.mysql:
                return parseRespCommon(res);
            case EnumClient.mysql2:
                return parseRespMysql2(res);
            default:
                console.warn(`[Kmore] Unsupported client value: '${clientType.toString()}' for getCurrentTime(). `);
                return '';
        }
    }
    else {
        throw new TypeError('Value of client is empty or not string');
    }
}
function parseRespCommon(res) {
    return res.rows[0].currenttime;
}
function parseRespMysql2(res) {
    // eslint-disable-next-line @typescript-eslint/no-unnecessary-condition, @typescript-eslint/prefer-optional-chain
    return res && res[0] && res[0][0] && res[0][0].currenttime
        ? res[0][0].currenttime
        : '';
}
export const defaultWrapIdentifierIgnoreRule = [
    'now()',
    'CURRENT_DATE',
    'CURRENT_TIME',
    'CURRENT_TIMESTAMP',
    // /\bCURRENT_DATE\b/u,
    // /\bCURRENT_TIMESTAMP\b/u,
    // /\bDATE_TRUNC(/u,
    // /\bINTERVAL\b/u,
];
/**
 * Convert identifier (field) to snake/camel case
 */
export function wrapIdentifier(value, origImpl, queryContext) {
    let ret = '';
    // do not convert if value is add by  builder.columns(columns)
    if (isIdentifierInColumns(value, queryContext?.columns)) {
        ret = origImpl(value);
        return ret;
    }
    if (value === '*') {
        return origImpl(value);
    }
    if (queryContext) {
        const { wrapIdentifierIgnoreRule } = queryContext;
        if (wrapIdentifierIgnoreRule?.length) {
            const flag = wrapIdentifierIgnoreRule.some((rule) => {
                if (typeof rule === 'string') {
                    return rule === value;
                }
                else if (rule instanceof RegExp) {
                    return rule.test(value);
                }
                else {
                    return false;
                }
            });
            if (flag) {
                return value;
            }
        }
        switch (queryContext.wrapIdentifierCaseConvert) {
            case CaseType.camel: {
                ret = origImpl(snakeToCamel(value));
                break;
            }
            case CaseType.snake: {
                ret = origImpl(camelToSnake(value));
                break;
            }
            case CaseType.pascal: {
                throw new TypeError('CaseType.pascal for wrapIdentifierCaseConvert not implemented yet');
            }
            default:
                ret = origImpl(value);
                break;
        }
    }
    else {
        ret = origImpl(value);
    }
    if (value === '' && ret === '``') {
        // fix for mysql when identifier is empty string
        // e.g. SELECT '' AS foo => SELECT `` AS foo, will be converted to SELECT '' AS foo
        ret = '\'\'';
    }
    return ret;
}
export function postProcessResponse(result, queryContext) {
    if (!queryContext) {
        return result;
    }
    const caseConvert = queryContext.postProcessResponseCaseConvert;
    // eslint-disable-next-line @typescript-eslint/no-unnecessary-condition
    if (!caseConvert) {
        return result;
    }
    switch (caseConvert) {
        case CaseType.camel:
            return postProcessResponseToCamel(result, queryContext);
        case CaseType.pascal:
            throw Error('Not implemented yet for pascal case conversion');
        // return postProcessResponseToPascal(result, queryContext)
        case CaseType.snake:
            return postProcessResponseToSnake(result, queryContext);
        default:
            return result;
    }
}
/**
 * Convert keys of result to camel case, if input is object
 */
export function postProcessResponseToCamel(result, _queryContext) {
    if (Array.isArray(result)) {
        const ret = result.map((row) => {
            const data = postProcessResponseToCamel(row, _queryContext);
            return data;
        });
        return ret;
    }
    else if (typeof result === 'object' && result) {
        const ret = _elementKeyToCamel(result, _queryContext);
        return ret;
    }
    return result;
}
function _elementKeyToCamel(row, _queryContext) {
    assert(typeof row === 'object' && row, 'row should be object');
    const columns = _queryContext?.columns;
    if (!columns) {
        const ret = genCamelKeysFrom(row);
        return ret;
    }
    const { resultNotConvertKeys, resultNeedConvertKeys } = genResultKeysData(row, columns);
    const results2 = genCamelKeysFrom(resultNeedConvertKeys);
    const ret = Object.assign(resultNotConvertKeys, results2);
    return ret;
}
/**
 * Convert keys of result to PascalCase, if input is object
 */
// export function postProcessResponseToPascal<T extends PostProcessInput = PostProcessInput>(
//   result: T,
//   _queryContext?: QueryContext,
// ): PostProcessRespRet<T, CaseType.pascal> {
//   if (Array.isArray(result)) {
//     const ret = result.map(row => postProcessResponseToCamel(row, _queryContext))
//     return ret as PostProcessRespRet<T, CaseType.pascal>
//   }
//   else if (typeof result === 'object' && result) {
//     const ret = genPascalKeysFrom(result)
//     return ret as PostProcessRespRet<T, CaseType.pascal>
//   }
//   return result as PostProcessRespRet<T, CaseType.pascal>
// }
/**
 * Convert keys of result to snake_case, if input is object
 */
export function postProcessResponseToSnake(result, _queryContext) {
    if (Array.isArray(result)) {
        const ret = result.map((row) => {
            const data = postProcessResponseToSnake(row, _queryContext);
            return data;
        });
        return ret;
    }
    else if (typeof result === 'object' && result) {
        const ret = _elementKeyToSnake(result, _queryContext);
        return ret;
    }
    return result;
}
function _elementKeyToSnake(row, _queryContext) {
    assert(typeof row === 'object' && row, 'row should be object');
    const columns = _queryContext?.columns;
    if (!columns) {
        const ret = genSnakeKeysFrom(row);
        return ret;
    }
    const { resultNotConvertKeys, resultNeedConvertKeys } = genResultKeysData(row, columns);
    const results2 = genSnakeKeysFrom(resultNeedConvertKeys);
    const ret = Object.assign(resultNotConvertKeys, results2);
    return ret;
}
function genResultKeysData(row, columns) {
    assert(typeof row === 'object' && row, 'row should be object');
    const resultNotConvertKeys = {};
    const resultNeedConvertKeys = {};
    Object.entries(row).forEach(([key, value]) => {
        // do not convert if value is add by  builder.columns(columns)
        if (isIdentifierInColumns(key, columns)) {
            Object.defineProperty(resultNotConvertKeys, key, {
                configurable: true,
                enumerable: true,
                writable: true,
                value,
            });
        }
        else {
            Object.defineProperty(resultNeedConvertKeys, key, {
                configurable: true,
                enumerable: true,
                writable: true,
                value,
            });
        }
    });
    return {
        resultNotConvertKeys,
        resultNeedConvertKeys,
    };
}
export function genCamelKeysFrom(input) {
    return camelKeys(input);
}
// function genPascalKeysFrom<From extends PostProcessRecord>(
//   input: From,
// ): RecordPascalKeys<From, '_'> {
//   return pascalCase(input)
// }
export function genSnakeKeysFrom(input) {
    return snakeKeys(input);
}
/**
 * @description only one level
 */
export function mergeDoWithInitData(initDoData, input) {
    if (typeof initDoData !== 'object' || Array.isArray(initDoData)) {
        throw new TypeError('initData not object');
    }
    const ret = {
        ...initDoData,
    };
    if (!input || typeof input !== 'object' || !Object.keys(input).length) {
        return ret;
    }
    Object.keys(ret).forEach((key) => {
        // @ts-ignore
        if (Object.hasOwn(input, key) && typeof input[key] !== 'undefined') {
            Object.defineProperty(ret, key, {
                configurable: true,
                enumerable: true,
                writable: true,
                // @ts-ignore
                value: input[key],
            });
        }
    });
    return ret;
}
function isIdentifierInColumns(value, columns) {
    if (!columns?.length) {
        return false;
    }
    const found = columns.some((row) => {
        if (row['fromSmartJoin']) {
            return false;
        }
        if (typeof row[value] !== 'undefined' && row[value] !== value) {
            return true;
        }
        return false;
    });
    return found;
}
//# sourceMappingURL=helper.js.map